@@ -1,14 +1,11 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import abstract_arrays, core
-from jax.core import Primitive, Tracer, Token
-from jax.lax import create_token
-
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
+from jax.core import ShapedArray
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -17,12 +14,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
+
+from mpi4jax._src.xla_bridge.device_descriptors import build_alltoall_descriptor
 
 
 # The Jax primitive
@@ -33,13 +44,12 @@ mpi_alltoall_impl = default_primitive_impl(mpi_alltoall_p)
 # This function applies the primitive to an AST
 @enforce_types(
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
 def alltoall(
     x,
     *,
     comm=None,
-    token=None,
+    token=NOTSET,
 ):
     """Perform an alltoall operation.
 
@@ -47,17 +57,12 @@ def alltoall(
         x: Array input to send. First axis must have size ``nproc``.
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
 
     Returns:
-        Tuple[DeviceArray, Token]:
-            - Received data.
-            - A new, modified token, that depends on this operation.
+        DeviceArray: Received data.
 
     """
-    if token is None:
-        token = create_token(x)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
@@ -68,18 +73,15 @@ def alltoall(
 
     comm = wrap_as_hashable(comm)
 
-    return tuple(
-        mpi_alltoall_p.bind(
-            x,
-            token,
-            comm=comm,
-        )
+    return mpi_alltoall_p.bind(
+        x,
+        comm=comm,
     )
 
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_alltoall_xla_encode_cpu(ctx, x, token, comm):
+def mpi_alltoall_xla_encode_cpu(ctx, x, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -97,9 +99,11 @@ def mpi_alltoall_xla_encode_cpu(ctx, x, token, comm):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         as_mhlo_constant(nitems_per_proc, _np.intc),
         x,
@@ -112,9 +116,9 @@ def mpi_alltoall_xla_encode_cpu(ctx, x, token, comm):
         token,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_alltoall",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         # force c order because first axis is special
         operand_layouts=get_default_layouts(operands, order="c"),
@@ -122,11 +126,14 @@ def mpi_alltoall_xla_encode_cpu(ctx, x, token, comm):
         has_side_effect=True,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-@translation_rule_gpu
-def mpi_alltoall_xla_encode_gpu(ctx, x, token, comm):
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_alltoall_descriptor
+    return results
 
+
+def mpi_alltoall_xla_encode_device(ctx, x, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -144,9 +151,11 @@ def mpi_alltoall_xla_encode_gpu(ctx, x, token, comm):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         x,
         token,
@@ -162,9 +171,9 @@ def mpi_alltoall_xla_encode_gpu(ctx, x, token, comm):
         to_mpi_handle(comm),
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_alltoall",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         # force c order because first axis is special
         operand_layouts=get_default_layouts(operands, order="c"),
@@ -173,19 +182,25 @@ def mpi_alltoall_xla_encode_gpu(ctx, x, token, comm):
         backend_config=descriptor,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+    return results
+
+
+mpi_alltoall_xla_encode_xpu = translation_rule_xpu(mpi_alltoall_xla_encode_device)
+mpi_alltoall_xla_encode_cuda = translation_rule_cuda(mpi_alltoall_xla_encode_device)
+
 
 # This function evaluates only the shapes during AST construction
-def mpi_alltoall_abstract_eval(xs, token, comm):
-    return (
-        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
-        core.abstract_token,
-    ), {effect}
+def mpi_alltoall_abstract_eval(xs, comm):
+    return ShapedArray(xs.shape, xs.dtype), {ordered_effect}
 
 
-mpi_alltoall_p.multiple_results = True
 mpi_alltoall_p.def_impl(mpi_alltoall_impl)
 mpi_alltoall_p.def_effectful_abstract_eval(mpi_alltoall_abstract_eval)
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_alltoall_p, mpi_alltoall_xla_encode_xpu, platform="xpu")
