@@ -1,14 +1,12 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import abstract_arrays, core
-from jax.core import Primitive, Tracer, Token
-from jax.lax import create_token
 
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
+from jax.core import ShapedArray
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -17,13 +15,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
 
+from mpi4jax._src.xla_bridge.device_descriptors import build_gather_descriptor
 
 # The Jax primitive
 mpi_gather_p = Primitive("gather_mpi")  # Create the primitive
@@ -34,14 +45,13 @@ mpi_gather_impl = default_primitive_impl(mpi_gather_p)
 @enforce_types(
     root=(_np.integer),
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
 def gather(
     x,
     root,
     *,
     comm=None,
-    token=None,
+    token=NOTSET,
 ):
     """Perform a gather operation.
 
@@ -60,16 +70,11 @@ def gather(
         root (int): Rank of the root MPI process.
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
 
     Returns:
-        Tuple[DeviceArray, Token]:
-            - Received data on root process, otherwise unmodified input.
-            - A new, modified token, that depends on this operation.
+        DeviceArray: Received data on root process, otherwise unmodified input.
     """
-    if token is None:
-        token = create_token(x)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
@@ -77,22 +82,21 @@ def gather(
     rank = comm.Get_rank()
     comm = wrap_as_hashable(comm)
 
-    res, token = mpi_gather_p.bind(
+    res = mpi_gather_p.bind(
         x,
-        token,
         root=root,
         comm=comm,
     )
 
     if rank != root:
-        return (x, token)
+        return x
 
-    return (res, token)
+    return res
 
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_gather_xla_encode_cpu(ctx, x, token, root, comm):
+def mpi_gather_xla_encode_cpu(ctx, x, root, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -117,9 +121,11 @@ def mpi_gather_xla_encode_cpu(ctx, x, token, root, comm):
 
     out_types = [
         ir.RankedTensorType.get(out_shape, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         as_mhlo_constant(nitems, _np.intc),
         x,
@@ -133,9 +139,9 @@ def mpi_gather_xla_encode_cpu(ctx, x, token, root, comm):
         token,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_gather",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         # enforce c order because the first axis is special
         operand_layouts=get_default_layouts(operands, order="c"),
@@ -143,11 +149,14 @@ def mpi_gather_xla_encode_cpu(ctx, x, token, root, comm):
         has_side_effect=True,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-@translation_rule_gpu
-def mpi_gather_xla_encode_gpu(ctx, x, token, root, comm):
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_gather_descriptor
+    return results
 
+
+def mpi_gather_xla_encode_device(ctx, x, root, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -172,9 +181,11 @@ def mpi_gather_xla_encode_gpu(ctx, x, token, root, comm):
 
     out_types = [
         ir.RankedTensorType.get(out_shape, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         x,
         token,
@@ -191,9 +202,9 @@ def mpi_gather_xla_encode_gpu(ctx, x, token, root, comm):
         to_mpi_handle(comm),
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_gather",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         # enforce c order because the first axis is special
         operand_layouts=get_default_layouts(operands, order="c"),
@@ -202,9 +213,19 @@ def mpi_gather_xla_encode_gpu(ctx, x, token, root, comm):
         backend_config=descriptor,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+
+    return results
+
+
+mpi_gather_xla_encode_xpu = translation_rule_xpu(mpi_gather_xla_encode_device)
+mpi_gather_xla_encode_cuda = translation_rule_cuda(mpi_gather_xla_encode_device)
+
 
 # This function evaluates only the shapes during AST construction
-def mpi_gather_abstract_eval(x, token, root, comm):
+def mpi_gather_abstract_eval(x, root, comm):
     comm = unpack_hashable(comm)
     rank = comm.Get_rank()
     size = comm.Get_size()
@@ -214,16 +235,13 @@ def mpi_gather_abstract_eval(x, token, root, comm):
     else:
         out_shape = (0,)
 
-    return (
-        abstract_arrays.ShapedArray(out_shape, x.dtype),
-        core.abstract_token,
-    ), {effect}
+    return ShapedArray(out_shape, x.dtype), {ordered_effect}
 
 
-mpi_gather_p.multiple_results = True
 mpi_gather_p.def_impl(mpi_gather_impl)
 mpi_gather_p.def_effectful_abstract_eval(mpi_gather_abstract_eval)
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_gather_p, mpi_gather_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_gather_p, mpi_gather_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_gather_p, mpi_gather_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_gather_p, mpi_gather_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_gather_p, mpi_gather_xla_encode_xpu, platform="xpu")
