@@ -1,14 +1,12 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import abstract_arrays, core
-from jax.core import Primitive, Tracer, Token
-from jax.lax import create_token
 
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
+from jax.core import ShapedArray
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -18,12 +16,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
+
+from mpi4jax._src.xla_bridge.device_descriptors import build_recv_descriptor
 
 
 # The Jax primitive
@@ -37,7 +49,6 @@ mpi_recv_impl = default_primitive_impl(mpi_recv_p)
     tag=_np.integer,
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
     status=(type(None), _MPI.Status, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
 def recv(
     x,
@@ -46,7 +57,7 @@ def recv(
     tag=_MPI.ANY_TAG,
     comm=None,
     status=None,
-    token=None,
+    token=NOTSET,
 ):
     """Perform a recv (receive) operation.
 
@@ -62,17 +73,11 @@ def recv(
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
         status (mpi4py.MPI.Status): Status object, can be used for introspection.
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
 
     Returns:
-        Tuple[DeviceArray, Token]:
-            - Received data.
-            - A new, modified token, that depends on this operation.
-
+        DeviceArray: Received data.
     """
-    if token is None:
-        token = create_token(x)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
@@ -82,15 +87,13 @@ def recv(
     if status is not None:
         status = wrap_as_hashable(status)
 
-    return tuple(
-        mpi_recv_p.bind(x, token, source=source, tag=tag, comm=comm, status=status)
-    )
+    return mpi_recv_p.bind(x, source=source, tag=tag, comm=comm, status=status)
 
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_recv_xla_encode_cpu(ctx, x, token, source, tag, comm, status):
-    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
+def mpi_recv_xla_encode_cpu(ctx, x, source, tag, comm, status):
+    from mpi4jax._src.xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
 
     comm = unpack_hashable(comm)
     status = unpack_hashable(status)
@@ -108,7 +111,7 @@ def mpi_recv_xla_encode_cpu(ctx, x, token, source, tag, comm, status):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
     if status is None:
@@ -116,6 +119,8 @@ def mpi_recv_xla_encode_cpu(ctx, x, token, source, tag, comm, status):
     else:
         status_ptr = to_mpi_ptr(status)
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         as_mhlo_constant(nitems, _np.intc),
         as_mhlo_constant(source, _np.intc),
@@ -126,20 +131,24 @@ def mpi_recv_xla_encode_cpu(ctx, x, token, source, tag, comm, status):
         token,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_recv",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-@translation_rule_gpu
-def mpi_recv_xla_encode_gpu(ctx, x, token, source, tag, comm, status):
-    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_recv_descriptor
+    return results
+
+
+def mpi_recv_xla_encode_device(ctx, x, source, tag, comm, status):
+    from mpi4jax._src.xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
 
     comm = unpack_hashable(comm)
     status = unpack_hashable(status)
@@ -157,7 +166,7 @@ def mpi_recv_xla_encode_gpu(ctx, x, token, source, tag, comm, status):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
     if status is None:
@@ -165,6 +174,8 @@ def mpi_recv_xla_encode_gpu(ctx, x, token, source, tag, comm, status):
     else:
         status_ptr = to_mpi_ptr(status)
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (token,)
 
     descriptor = build_recv_descriptor(
@@ -176,9 +187,9 @@ def mpi_recv_xla_encode_gpu(ctx, x, token, source, tag, comm, status):
         status_ptr,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_recv",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
@@ -186,19 +197,26 @@ def mpi_recv_xla_encode_gpu(ctx, x, token, source, tag, comm, status):
         backend_config=descriptor,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+
+    return results
+
+
+mpi_recv_xla_encode_xpu = translation_rule_xpu(mpi_recv_xla_encode_device)
+mpi_recv_xla_encode_cuda = translation_rule_cuda(mpi_recv_xla_encode_device)
+
 
 # This function evaluates only the shapes during AST construction
-def mpi_recv_abstract_eval(xs, token, source, tag, comm, status):
-    return (
-        abstract_arrays.ShapedArray(xs.shape, xs.dtype),
-        core.abstract_token,
-    ), {effect}
+def mpi_recv_abstract_eval(xs, source, tag, comm, status):
+    return ShapedArray(xs.shape, xs.dtype), {ordered_effect}
 
 
-mpi_recv_p.multiple_results = True
 mpi_recv_p.def_impl(mpi_recv_impl)
 mpi_recv_p.def_effectful_abstract_eval(mpi_recv_abstract_eval)
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_recv_p, mpi_recv_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_recv_p, mpi_recv_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_recv_p, mpi_recv_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_recv_p, mpi_recv_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_recv_p, mpi_recv_xla_encode_xpu, platform="xpu")
