@@ -1,14 +1,12 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import abstract_arrays, core
-from jax.core import Primitive, Tracer, Token
-from jax.lax import create_token
 
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
+from jax.core import ShapedArray
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -17,13 +15,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
 
+from mpi4jax._src.xla_bridge.device_descriptors import build_reduce_descriptor
 
 # The Jax primitive
 mpi_reduce_p = Primitive("reduce_mpi")  # Create the primitive
@@ -35,9 +46,8 @@ mpi_reduce_impl = default_primitive_impl(mpi_reduce_p)
     op=(_MPI.Op, HashableMPIType),
     root=(_np.integer),
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
-def reduce(x, op, root, *, comm=None, token=None):
+def reduce(x, op, root, *, comm=None, token=NOTSET):
     """Perform a reduce operation.
 
     Arguments:
@@ -46,18 +56,12 @@ def reduce(x, op, root, *, comm=None, token=None):
         root (int): Rank of the root MPI process.
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
 
     Returns:
-        Tuple[DeviceArray, Token]:
-            - Result of the reduce operation on root process, otherwise
-              unmodified input.
-            - A new, modified token, that depends on this operation.
-
+        DeviceArray: Result of the reduce operation on root process, otherwise
+            unmodified input.
     """
-    if token is None:
-        token = create_token(x)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
@@ -66,17 +70,17 @@ def reduce(x, op, root, *, comm=None, token=None):
 
     op = wrap_as_hashable(op)
     comm = wrap_as_hashable(comm)
-    res, token = mpi_reduce_p.bind(x, token, op=op, root=root, comm=comm)
+    res = mpi_reduce_p.bind(x, op=op, root=root, comm=comm)
 
     if rank != root:
-        return (x, token)
+        return x
 
-    return (res, token)
+    return res
 
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_reduce_xla_encode_cpu(ctx, x, token, op, root, comm):
+def mpi_reduce_xla_encode_cpu(ctx, x, op, root, comm):
     op = unpack_hashable(op)
     comm = unpack_hashable(comm)
 
@@ -100,9 +104,11 @@ def mpi_reduce_xla_encode_cpu(ctx, x, token, op, root, comm):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         as_mhlo_constant(nitems, _np.intc),
         x,
@@ -113,20 +119,23 @@ def mpi_reduce_xla_encode_cpu(ctx, x, token, op, root, comm):
         token,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_reduce",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-@translation_rule_gpu
-def mpi_reduce_xla_encode_gpu(ctx, x, token, op, root, comm):
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_reduce_descriptor
+    return results
 
+
+def mpi_reduce_xla_encode_device(ctx, x, op, root, comm):
     op = unpack_hashable(op)
     comm = unpack_hashable(comm)
 
@@ -150,9 +159,11 @@ def mpi_reduce_xla_encode_gpu(ctx, x, token, op, root, comm):
 
     out_types = [
         ir.RankedTensorType.get(dims, dtype),
-        *token_type(),
+        token_type(),
     ]
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         x,
         token,
@@ -166,9 +177,9 @@ def mpi_reduce_xla_encode_gpu(ctx, x, token, op, root, comm):
         dtype_handle,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_reduce",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
@@ -176,9 +187,19 @@ def mpi_reduce_xla_encode_gpu(ctx, x, token, op, root, comm):
         backend_config=descriptor,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+
+    return results
+
+
+mpi_reduce_xla_encode_xpu = translation_rule_xpu(mpi_reduce_xla_encode_device)
+mpi_reduce_xla_encode_cuda = translation_rule_cuda(mpi_reduce_xla_encode_device)
+
 
 # This function evaluates only the shapes during AST construction
-def mpi_reduce_abstract_eval(xs, token, op, root, comm):
+def mpi_reduce_abstract_eval(xs, op, root, comm):
     comm = unpack_hashable(comm)
     rank = comm.Get_rank()
 
@@ -187,16 +208,13 @@ def mpi_reduce_abstract_eval(xs, token, op, root, comm):
     else:
         dims = xs.shape
 
-    return (
-        abstract_arrays.ShapedArray(dims, xs.dtype),
-        core.abstract_token,
-    ), {effect}
+    return ShapedArray(dims, xs.dtype), {ordered_effect}
 
 
-mpi_reduce_p.multiple_results = True
 mpi_reduce_p.def_impl(mpi_reduce_impl)
 mpi_reduce_p.def_effectful_abstract_eval(mpi_reduce_abstract_eval)
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_reduce_p, mpi_reduce_xla_encode_xpu, platform="xpu")
