@@ -1,14 +1,11 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import core
-from jax.core import Primitive, Tracer, Token
-from jax.lax import create_token
 
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -17,13 +14,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
 
+from mpi4jax._src.xla_bridge.device_descriptors import build_send_descriptor
 
 # The Jax primitive
 mpi_send_p = Primitive("send_mpi")  # Create the primitive
@@ -35,9 +45,15 @@ mpi_send_impl = default_primitive_impl(mpi_send_p)
     dest=_np.integer,
     tag=_np.integer,
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
-def send(x, dest, *, tag=0, comm=None, token=None):
+def send(
+    x,
+    dest,
+    *,
+    tag=0,
+    comm=None,
+    token=NOTSET,
+):
     """Perform a send operation.
 
     Arguments:
@@ -46,26 +62,20 @@ def send(x, dest, *, tag=0, comm=None, token=None):
         tag (int): Tag of this message.
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
-
-    Returns:
-        Token: A new, modified token, that depends on this operation.
 
     """
-    if token is None:
-        token = create_token(x)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
 
     comm = wrap_as_hashable(comm)
-    return mpi_send_p.bind(x, token, dest=dest, tag=tag, comm=comm)
+    return mpi_send_p.bind(x, dest=dest, tag=tag, comm=comm)
 
 
 # This function compiles the operation
 @translation_rule_cpu
-def mpi_send_xla_encode_cpu(ctx, x, token, dest, tag, comm):
+def mpi_send_xla_encode_cpu(ctx, x, dest, tag, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -78,7 +88,9 @@ def mpi_send_xla_encode_cpu(ctx, x, token, dest, tag, comm):
     nitems = _np.prod(dims, dtype=int)
     dtype_handle = to_dtype_handle(x_nptype)
 
-    out_types = token_type()
+    out_types = [token_type()]
+
+    token = get_token_effect(ctx, ordered_effect)
 
     operands = (
         as_mhlo_constant(nitems, _np.intc),
@@ -90,23 +102,23 @@ def mpi_send_xla_encode_cpu(ctx, x, token, dest, tag, comm):
         token,
     )
 
-    # JAX insists on outputs being iterable
-    return [
-        hlo_custom_call(
-            b"mpi_send",
-            out_types=out_types,
-            operands=operands,
-            operand_layouts=get_default_layouts(operands),
-            result_layouts=get_default_layouts(out_types),
-            has_side_effect=True,
-        )
-    ]
+    result_obj = custom_call(
+        b"mpi_send",
+        result_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
+        has_side_effect=True,
+    )
+
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+
+    return results
 
 
-@translation_rule_gpu
-def mpi_send_xla_encode_gpu(ctx, x, token, dest, tag, comm):
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_send_descriptor
-
+def mpi_send_xla_encode_device(ctx, x, dest, tag, comm):
     comm = unpack_hashable(comm)
 
     x_aval, *_ = ctx.avals_in
@@ -119,7 +131,9 @@ def mpi_send_xla_encode_gpu(ctx, x, token, dest, tag, comm):
     nitems = _np.prod(dims, dtype=int)
     dtype_handle = to_dtype_handle(x_nptype)
 
-    out_types = token_type()
+    out_types = [token_type()]
+
+    token = get_token_effect(ctx, ordered_effect)
 
     operands = (
         x,
@@ -134,28 +148,37 @@ def mpi_send_xla_encode_gpu(ctx, x, token, dest, tag, comm):
         dtype_handle,
     )
 
-    # JAX insists on outputs being iterable
-    return [
-        hlo_custom_call(
-            b"mpi_send",
-            out_types=out_types,
-            operands=operands,
-            operand_layouts=get_default_layouts(operands),
-            result_layouts=get_default_layouts(out_types),
-            has_side_effect=True,
-            backend_config=descriptor,
-        )
-    ]
+    result_obj = custom_call(
+        b"mpi_send",
+        result_types=out_types,
+        operands=operands,
+        operand_layouts=get_default_layouts(operands),
+        result_layouts=get_default_layouts(out_types),
+        has_side_effect=True,
+        backend_config=descriptor,
+    )
+
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
+
+    return results
+
+
+mpi_send_xla_encode_xpu = translation_rule_xpu(mpi_send_xla_encode_device)
+mpi_send_xla_encode_cuda = translation_rule_cuda(mpi_send_xla_encode_device)
 
 
 # This function evaluates only the shapes during AST construction
-def mpi_send_abstract_eval(xs, token, dest, tag, comm):
-    return core.abstract_token, {effect}
+def mpi_send_abstract_eval(xs, dest, tag, comm):
+    return (), {ordered_effect}
 
 
+mpi_send_p.multiple_results = True
 mpi_send_p.def_impl(mpi_send_impl)
 mpi_send_p.def_effectful_abstract_eval(mpi_send_abstract_eval)
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_send_p, mpi_send_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_send_p, mpi_send_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_send_p, mpi_send_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_send_p, mpi_send_xla_encode_xpu, platform="xpu")
