@@ -1,15 +1,15 @@
 import numpy as _np
 from mpi4py import MPI as _MPI
 
-from jax import abstract_arrays, core
-from jax.core import Primitive, Tracer, Token
+from jax.core import get_aval
 from jax.interpreters import ad, batching
-from jax.lax import create_token
+from jax.core import ShapedArray
+
 
-from jax.interpreters import mlir
 import jaxlib.mlir.ir as ir
+from jax.interpreters.mlir import custom_call
 
-from ..utils import (
+from mpi4jax._src.utils import (
     HashableMPIType,
     default_primitive_impl,
     to_dtype_handle,
@@ -19,12 +19,26 @@ from ..utils import (
     wrap_as_hashable,
     as_mhlo_constant,
     get_default_layouts,
-    effect,
+    ordered_effect,
+    NOTSET,
+    raise_if_token_is_set,
 )
-from ..jax_compat import hlo_custom_call, token_type
-from ..decorators import translation_rule_cpu, translation_rule_gpu
-from ..validation import enforce_types
-from ..comm import get_default_comm
+from mpi4jax._src.jax_compat import (
+    register_lowering,
+    token_type,
+    get_token_effect,
+    set_token_effect,
+    Primitive,
+)
+from mpi4jax._src.decorators import (
+    translation_rule_cpu,
+    translation_rule_cuda,
+    translation_rule_xpu,
+)
+from mpi4jax._src.validation import enforce_types
+from mpi4jax._src.comm import get_default_comm
+
+from mpi4jax._src.xla_bridge.device_descriptors import build_sendrecv_descriptor
 
 
 # The Jax primitive
@@ -40,7 +54,6 @@ mpi_sendrecv_impl = default_primitive_impl(mpi_sendrecv_p)
     recvtag=_np.integer,
     comm=(type(None), _MPI.Intracomm, HashableMPIType),
     status=(type(None), _MPI.Status, HashableMPIType),
-    token=(type(None), Token, Tracer),
 )
 def sendrecv(
     sendbuf,
@@ -52,7 +65,7 @@ def sendrecv(
     recvtag=_MPI.ANY_TAG,
     comm=None,
     status=None,
-    token=None,
+    token=NOTSET,
 ):
     """Perform a sendrecv operation.
 
@@ -71,17 +84,12 @@ def sendrecv(
         comm (mpi4py.MPI.Comm): The MPI communicator to use (defaults to
             a clone of :obj:`COMM_WORLD`).
         status (mpi4py.MPI.Status): Status object, can be used for introspection.
-        token (Token): XLA token to use to ensure correct execution order.
-            If not given, a new token is generated.
 
     Returns:
-        Tuple[DeviceArray, Token]:
-            - Received data.
-            - A new, modified token, that depends on this operation.
+        DeviceArray: Received data.
 
     """
-    if token is None:
-        token = create_token(sendbuf)
+    raise_if_token_is_set(token)
 
     if comm is None:
         comm = get_default_comm()
@@ -91,19 +99,16 @@ def sendrecv(
     if status is not None:
         status = wrap_as_hashable(status)
 
-    return tuple(
-        mpi_sendrecv_p.bind(
-            sendbuf,
-            recvbuf,
-            token,
-            source=source,
-            dest=dest,
-            sendtag=sendtag,
-            recvtag=recvtag,
-            comm=comm,
-            status=status,
-            _must_transpose=False,
-        )
+    return mpi_sendrecv_p.bind(
+        sendbuf,
+        recvbuf,
+        source=source,
+        dest=dest,
+        sendtag=sendtag,
+        recvtag=recvtag,
+        comm=comm,
+        status=status,
+        _must_transpose=False,
     )
 
 
@@ -113,7 +118,6 @@ def mpi_sendrecv_xla_encode_cpu(
     ctx,
     sendbuf,
     recvbuf,
-    token,
     source,
     dest,
     sendtag,
@@ -122,7 +126,7 @@ def mpi_sendrecv_xla_encode_cpu(
     status,
     _must_transpose=False,
 ):
-    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
+    from mpi4jax._src.xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
 
     # when performing forward diff, the gradient will follow the sent message.
     # so if you do a sendrecv from rank 0 to 1, the gradient wrt the inputs of rank 0
@@ -159,7 +163,7 @@ def mpi_sendrecv_xla_encode_cpu(
 
     out_types = [
         ir.RankedTensorType.get(recv_dims, recv_dtype),
-        *token_type(),
+        token_type(),
     ]
 
     if status is None:
@@ -167,6 +171,8 @@ def mpi_sendrecv_xla_encode_cpu(
     else:
         status_ptr = to_mpi_ptr(status)
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         as_mhlo_constant(send_nitems, _np.intc),
         sendbuf,
@@ -182,31 +188,34 @@ def mpi_sendrecv_xla_encode_cpu(
         token,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_sendrecv",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
         has_side_effect=True,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-@translation_rule_gpu
-def mpi_sendrecv_xla_encode_gpu(
+    return results
+
+
+def mpi_sendrecv_xla_encode_device(
     ctx,
     sendbuf,
     recvbuf,
-    token,
     source,
     dest,
     sendtag,
     recvtag,
     comm,
     status,
-    _must_transpose=False,
+    _must_transpose,
 ):
-
     if _must_transpose:
         raise RuntimeError(
             "sendrecv cannot be used with forward-mode (vjp) autodiff, because "
@@ -214,8 +223,7 @@ def mpi_sendrecv_xla_encode_gpu(
             "desired one. Use reverse-mode (jvp) differentiation instead."
         )
 
-    from ..xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
-    from ..xla_bridge.mpi_xla_bridge_gpu import build_sendrecv_descriptor
+    from mpi4jax._src.xla_bridge.mpi_xla_bridge import MPI_STATUS_IGNORE_ADDR
 
     comm = unpack_hashable(comm)
     status = unpack_hashable(status)
@@ -240,7 +248,7 @@ def mpi_sendrecv_xla_encode_gpu(
 
     out_types = [
         ir.RankedTensorType.get(recv_dims, recv_dtype),
-        *token_type(),
+        token_type(),
     ]
 
     if status is None:
@@ -248,6 +256,8 @@ def mpi_sendrecv_xla_encode_gpu(
     else:
         status_ptr = to_mpi_ptr(status)
 
+    token = get_token_effect(ctx, ordered_effect)
+
     operands = (
         sendbuf,
         token,
@@ -266,9 +276,9 @@ def mpi_sendrecv_xla_encode_gpu(
         status_ptr,
     )
 
-    return hlo_custom_call(
+    result_obj = custom_call(
         b"mpi_sendrecv",
-        out_types=out_types,
+        result_types=out_types,
         operands=operands,
         operand_layouts=get_default_layouts(operands),
         result_layouts=get_default_layouts(out_types),
@@ -276,12 +286,18 @@ def mpi_sendrecv_xla_encode_gpu(
         backend_config=descriptor,
     )
 
+    results = list(result_obj.results)
+    token = results.pop(-1)
+    set_token_effect(ctx, ordered_effect, token)
 
-# This function evaluates only the shapes during AST construction
-def mpi_sendrecv_abstract_eval(
+    return results
+
+
+@translation_rule_xpu
+def mpi_sendrecv_xla_encode_xpu(
+    ctx,
     sendbuf,
     recvbuf,
-    token,
     source,
     dest,
     sendtag,
@@ -290,10 +306,60 @@ def mpi_sendrecv_abstract_eval(
     status,
     _must_transpose=False,
 ):
-    return (
-        abstract_arrays.ShapedArray(recvbuf.shape, recvbuf.dtype),
-        core.abstract_token,
-    ), {effect}
+    return mpi_sendrecv_xla_encode_device(
+        ctx,
+        sendbuf,
+        recvbuf,
+        source,
+        dest,
+        sendtag,
+        recvtag,
+        comm,
+        status,
+        _must_transpose,
+    )
+
+
+@translation_rule_cuda
+def mpi_sendrecv_xla_encode_cuda(
+    ctx,
+    sendbuf,
+    recvbuf,
+    source,
+    dest,
+    sendtag,
+    recvtag,
+    comm,
+    status,
+    _must_transpose=False,
+):
+    return mpi_sendrecv_xla_encode_device(
+        ctx,
+        sendbuf,
+        recvbuf,
+        source,
+        dest,
+        sendtag,
+        recvtag,
+        comm,
+        status,
+        _must_transpose,
+    )
+
+
+# This function evaluates only the shapes during AST construction
+def mpi_sendrecv_abstract_eval(
+    sendbuf,
+    recvbuf,
+    source,
+    dest,
+    sendtag,
+    recvtag,
+    comm,
+    status,
+    _must_transpose=False,
+):
+    return ShapedArray(recvbuf.shape, recvbuf.dtype), {ordered_effect}
 
 
 def mpi_sendrecv_batch_eval(
@@ -307,15 +373,14 @@ def mpi_sendrecv_batch_eval(
     status,
     _must_transpose=False,
 ):
-
-    sendbuf, recvbuf, token = in_args
+    sendbuf, recvbuf = in_args
 
     assert batch_axes[0] == batch_axes[1]
+    ax = batch_axes[0]
 
     res = mpi_sendrecv_p.bind(
         sendbuf,
         recvbuf,
-        token,
         source=source,
         dest=dest,
         sendtag=sendtag,
@@ -324,7 +389,7 @@ def mpi_sendrecv_batch_eval(
         status=status,
         _must_transpose=_must_transpose,
     )
-    return res, (batch_axes[0], batch_axes[2])
+    return res, ax
 
 
 def mpi_sendrecv_value_and_jvp(
@@ -338,13 +403,12 @@ def mpi_sendrecv_value_and_jvp(
     status,
     _must_transpose=False,
 ):
-    sendbuf, recvbuf, token = in_args
-    send_tan, recv_tan, token_tan = tan_args
+    sendbuf, recvbuf = in_args
+    send_tan, recv_tan = tan_args
 
-    val, token = mpi_sendrecv_p.bind(
+    val = mpi_sendrecv_p.bind(
         sendbuf,
         recvbuf,
-        token,
         source=source,
         dest=dest,
         sendtag=sendtag,
@@ -354,11 +418,9 @@ def mpi_sendrecv_value_and_jvp(
         _must_transpose=_must_transpose,
     )
 
-    # throw away return token to work around jax#6285
-    jvp, token_jvp = mpi_sendrecv_p.bind(
+    jvp = mpi_sendrecv_p.bind(
         send_tan,
         recv_tan,
-        token,
         source=source,
         dest=dest,
         sendtag=sendtag,
@@ -368,20 +430,16 @@ def mpi_sendrecv_value_and_jvp(
         _must_transpose=not _must_transpose,
     )
 
-    return (val, token), (jvp, ad.Zero.from_value(token_jvp))
+    return val, jvp
 
 
 def mpi_sendrecv_transpose_rule(
-    tan_args, *x_args, source, dest, sendtag, recvtag, comm, status, _must_transpose
+    out_tan, *x_args, source, dest, sendtag, recvtag, comm, status, _must_transpose
 ):
-    _, _, token = x_args
-    out_tan, token_tan = tan_args
-
     # swap the sender and receiver
-    res, token = mpi_sendrecv_p.bind(
+    res = mpi_sendrecv_p.bind(
         out_tan,
         out_tan,
-        token,
         source=dest,
         dest=source,
         sendtag=sendtag,
@@ -390,10 +448,9 @@ def mpi_sendrecv_transpose_rule(
         status=status,
         _must_transpose=not _must_transpose,
     )
-    return res, ad.Zero.from_value(res), token_tan
+    return (res, ad.Zero(get_aval(res)))
 
 
-mpi_sendrecv_p.multiple_results = True
 mpi_sendrecv_p.def_impl(mpi_sendrecv_impl)
 mpi_sendrecv_p.def_effectful_abstract_eval(mpi_sendrecv_abstract_eval)
 
@@ -403,5 +460,6 @@ ad.primitive_jvps[mpi_sendrecv_p] = mpi_sendrecv_value_and_jvp
 ad.primitive_transposes[mpi_sendrecv_p] = mpi_sendrecv_transpose_rule
 
 # assign to the primitive the correct encoder
-mlir.register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_cpu, platform="cpu")
-mlir.register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_gpu, platform="cuda")
+register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_cpu, platform="cpu")
+register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_cuda, platform="cuda")
+register_lowering(mpi_sendrecv_p, mpi_sendrecv_xla_encode_xpu, platform="xpu")
