@@ -1,4 +1,5 @@
 import functools
+import hashlib
 
 from mpi4py import MPI as _MPI
 
@@ -8,16 +9,44 @@ from jax.interpreters import xla, mlir
 import jaxlib.mlir.ir as ir
 from jaxlib.mlir.dialects import mhlo
 
-from .jax_compat import token_type, register_effect
+from .jax_compat import token_type, register_effect, EffectType
 
 
-class MPIEffect:
+# Sentinel value for default arguments
+NOTSET = object()
+
+
+def raise_if_token_is_set(token):
+    """Raises an error if the token is set."""
+    if token is not NOTSET:
+        raise RuntimeError(
+            "Explicit token management is not supported for mpi4jax>=0.8.0. "
+            "Tokens are now managed automatically and must not be passed "
+            "as arguments to collective operations anymore.\n"
+            "That is, please adjust your code like this:\n"
+            "     # For mpi4jax<0.8.0:\n"
+            "     result, token = mpi4jax.allgather(x, token=token)\n"
+            "     # For mpi4jax>=0.8.0:\n"
+            "     result = mpi4jax.allgather(x)"
+        )
+
+
+class MPIEffect(EffectType):
     def __hash__(self):
         # enforce a constant (known) hash
-        return hash("I love mpi4jax")
+        return int(hashlib.md5("I love mpi4jax".encode("utf-8")).hexdigest(), 16)
+
+
+class OrderedMPIEffect(EffectType):
+    def __hash__(self):
+        # enforce a constant (known) hash
+        return int(
+            hashlib.md5("I love mpi4jax very much".encode("utf-8")).hexdigest(), 16
+        )
 
 
 effect = register_effect(MPIEffect)
+ordered_effect = register_effect(OrderedMPIEffect, ordered=True)
 
 
 def default_primitive_impl(primitive):
@@ -36,7 +65,7 @@ def as_mhlo_constant(val, dtype):
 
 
 def get_default_layouts(operands, order="c"):
-    (token,) = token_type()
+    token = token_type()
     layouts = []
 
     if order == "c":
@@ -44,7 +73,7 @@ def get_default_layouts(operands, order="c"):
     elif order == "f":
         default_layout = lambda t: tuple(range(len(t.shape)))
     else:
-        raise ValueError("Unknown order: {}".format(order))
+        raise ValueError(f"Unknown order: {order}")
 
     for op in operands:
         if isinstance(op, (ir.Value)):
@@ -61,7 +90,7 @@ def get_default_layouts(operands, order="c"):
             layouts.append(())
 
         else:
-            raise ValueError("Unknown operand type: {}".format(type(op)))
+            raise ValueError(f"Unknown operand type: {type(op)}")
 
     return layouts
 
@@ -110,7 +139,7 @@ def to_dtype_handle(dtype):
     """
     dtype_name = _np.dtype(dtype).name
     if dtype_name not in MPI_TYPE_MAP:
-        raise RuntimeError("Unknown MPI type for dtype {}".format(dtype_name))
+        raise RuntimeError(f"Unknown MPI type for dtype {dtype_name}")
 
     mpi_type = getattr(_MPI, MPI_TYPE_MAP[dtype_name])
     return to_mpi_handle(mpi_type)
@@ -150,4 +179,13 @@ def has_cuda_support() -> bool:
     """
     from . import xla_bridge
 
-    return xla_bridge.HAS_GPU_EXT
+    return xla_bridge.HAS_CUDA_EXT
+
+
+def has_sycl_support() -> bool:
+    """Returns True if mpi4jax is built with SYCL support and can be used with XPU-based
+    jax-arrays, False otherwise.
+    """
+    from . import xla_bridge
+
+    return xla_bridge.HAS_XPU_EXT
