@@ -13,7 +13,7 @@ def test_allgather():
 
     arr = jnp.ones((3, 2)) * rank
 
-    res, _ = allgather(arr)
+    res = allgather(arr)
     for p in range(size):
         assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)
 
@@ -23,7 +23,7 @@ def test_allgather_jit():
 
     arr = jnp.ones((3, 2)) * rank
 
-    res = jax.jit(lambda x: allgather(x)[0])(arr)
+    res = jax.jit(lambda x: allgather(x))(arr)
     for p in range(size):
         assert jnp.array_equal(res[p], jnp.ones((3, 2)) * p)
 
@@ -32,7 +32,7 @@ def test_allgather_scalar():
     from mpi4jax import allgather
 
     arr = rank
-    res, _ = allgather(arr)
+    res = allgather(arr)
     assert jnp.array_equal(res, jnp.arange(size))
 
 
@@ -40,5 +40,5 @@ def test_allgather_scalar_jit():
     from mpi4jax import allgather
 
     arr = rank
-    res = jax.jit(lambda x: allgather(x)[0])(arr)
+    res = jax.jit(lambda x: allgather(x))(arr)
     assert jnp.array_equal(res, jnp.arange(size))
