@@ -16,7 +16,7 @@ def test_allreduce():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res, token = allreduce(arr, op=MPI.SUM)
+    res = allreduce(arr, op=MPI.SUM)
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -27,7 +27,7 @@ def test_allreduce_jit():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
+    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM))(arr)
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -38,7 +38,7 @@ def test_allreduce_scalar():
     arr = 1
     _arr = 1
 
-    res, token = allreduce(arr, op=MPI.SUM)
+    res = allreduce(arr, op=MPI.SUM)
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -49,7 +49,7 @@ def test_allreduce_scalar_jit():
     arr = 1
     _arr = 1
 
-    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM)[0])(arr)
+    res = jax.jit(lambda x: allreduce(x, op=MPI.SUM))(arr)
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -60,7 +60,7 @@ def test_allreduce_vmap():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res = jax.vmap(lambda x: allreduce(x, op=MPI.SUM)[0], in_axes=0, out_axes=0)(arr)
+    res = jax.vmap(lambda x: allreduce(x, op=MPI.SUM), in_axes=0, out_axes=0)(arr)
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -71,9 +71,9 @@ def test_allreduce_vmap_jit():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res = jax.jit(
-        jax.vmap(lambda x: allreduce(x, op=MPI.SUM)[0], in_axes=0, out_axes=0)
-    )(arr)
+    res = jax.jit(jax.vmap(lambda x: allreduce(x, op=MPI.SUM), in_axes=0, out_axes=0))(
+        arr
+    )
     assert jnp.array_equal(res, arr * size)
     assert jnp.array_equal(_arr, arr)
 
@@ -84,7 +84,7 @@ def test_allreduce_transpose():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(_arr)
+    (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM), arr)(_arr)
     assert jnp.array_equal(_arr, res)
 
 
@@ -95,7 +95,7 @@ def test_allreduce_transpose_jit():
     _arr = arr.copy()
 
     def f(x):
-        (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(x)
+        (res,) = jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM), arr)(x)
         return res
 
     res = jax.jit(f)(arr)
@@ -111,10 +111,10 @@ def test_allreduce_transpose2():
     _arr2 = arr.copy()
 
     def lt(y):
-        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(y)[0]
+        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM), arr)(y)[0]
 
     (res,) = jax.linear_transpose(lt, _arr)(_arr2)
-    expected, _ = allreduce(_arr2, op=MPI.SUM)
+    expected = allreduce(_arr2, op=MPI.SUM)
     assert jnp.array_equal(expected, res)
 
 
@@ -127,14 +127,14 @@ def test_allreduce_transpose2_jit():
     _arr2 = arr.copy()
 
     def lt(y):
-        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM)[0], arr)(y)[0]
+        return jax.linear_transpose(lambda x: allreduce(x, op=MPI.SUM), arr)(y)[0]
 
     def f(x):
         (res,) = jax.linear_transpose(lt, _arr)(x)
         return res
 
     res = jax.jit(f)(_arr2)
-    expected, _ = allreduce(_arr2, op=MPI.SUM)
+    expected = allreduce(_arr2, op=MPI.SUM)
     assert jnp.array_equal(expected, res)
 
 
@@ -144,20 +144,20 @@ def test_allreduce_grad():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res, grad = jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM)[0].sum())(arr)
+    res, grad = jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM).sum())(arr)
     assert jnp.array_equal(res, arr.sum() * size)
     assert jnp.array_equal(_arr, grad)
 
-    res, grad = jax.jit(
-        jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM)[0].sum())
-    )(arr)
+    res, grad = jax.jit(jax.value_and_grad(lambda x: allreduce(x, op=MPI.SUM).sum()))(
+        arr
+    )
     assert jnp.array_equal(res, arr.sum() * size)
     assert jnp.array_equal(_arr, grad)
 
     def testfun(x):
-        y, token = allreduce(x, op=MPI.SUM)
+        y = allreduce(x, op=MPI.SUM)
         z = x + 2 * y  # noqa: F841
-        res, token = allreduce(x, op=MPI.SUM, token=token)
+        res = allreduce(x, op=MPI.SUM)
         return res.sum()
 
     res, grad = jax.jit(jax.value_and_grad(testfun))(arr)
@@ -171,11 +171,11 @@ def test_allreduce_jvp():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res, jvp = jax.jvp(lambda x: allreduce(x, op=MPI.SUM)[0], (arr,), (_arr,))
+    res, jvp = jax.jvp(lambda x: allreduce(x, op=MPI.SUM), (arr,), (_arr,))
 
-    expected, _ = allreduce(arr, op=MPI.SUM)
+    expected = allreduce(arr, op=MPI.SUM)
     assert jnp.array_equal(expected, res)
-    expected, _ = allreduce(_arr, op=MPI.SUM)
+    expected = allreduce(_arr, op=MPI.SUM)
     assert jnp.array_equal(expected, jvp)
 
 
@@ -185,10 +185,10 @@ def test_allreduce_vjp():
     arr = jnp.ones((3, 2))
     _arr = arr.copy()
 
-    res, vjp_fun = jax.vjp(lambda x: allreduce(x, op=MPI.SUM)[0], arr)
+    res, vjp_fun = jax.vjp(lambda x: allreduce(x, op=MPI.SUM), arr)
     (vjp,) = vjp_fun(_arr)
 
-    expected, _ = allreduce(arr, op=MPI.SUM)
+    expected = allreduce(arr, op=MPI.SUM)
     assert jnp.array_equal(expected, res)
     assert jnp.array_equal(_arr, vjp)
 
@@ -197,9 +197,10 @@ def test_allreduce_chained():
     from mpi4jax import allreduce
 
     def foo(x):
-        token = jax.lax.create_token()
-        x1, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
-        x2, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
+        x1 = allreduce(x, op=MPI.SUM, comm=comm)
+        print(x1)
+        x2 = allreduce(x, op=MPI.SUM, comm=comm)
+        print(x2)
         return x1 + x2
 
     res_t = jax.grad(foo)(0.0)
@@ -212,9 +213,8 @@ def test_allreduce_chained_jit():
     from mpi4jax import allreduce
 
     def foo(x):
-        token = jax.lax.create_token()
-        x1, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
-        x2, token = allreduce(x, op=MPI.SUM, comm=comm, token=token)
+        x1 = allreduce(x, op=MPI.SUM, comm=comm)
+        x2 = allreduce(x, op=MPI.SUM, comm=comm)
         return x1 + x2
 
     res_t = jax.jit(jax.grad(foo))(0.0)
@@ -231,14 +231,14 @@ def test_custom_vjp():
     def f(x, y):
         r = jnp.sin(x) * y
         r = r.sum()
-        return allreduce(r, op=MPI.SUM)[0]
+        return allreduce(r, op=MPI.SUM)
 
     def f_fwd(x, y):
         # Returns primal output and residuals to be used in backward pass by f_bwd.
         return f(x, y), (jnp.cos(x), jnp.sin(x), y)
 
     def f_bwd(res, g):
-        g = allreduce(g, op=MPI.SUM)[0]
+        g = allreduce(g, op=MPI.SUM)
         cos_x, sin_x, y = res  # Gets residuals computed in f_fwd
         return (cos_x * g * y, sin_x * g)
 
@@ -276,12 +276,12 @@ def test_advanced_jvp():
     @partial(jax.custom_vjp, nondiff_argnums=(0, 1, 2))
     def _expect(n_chains, log_pdf, expected_fun, pars, x, *expected_fun_args):
         L_x = expected_fun(pars, x, *expected_fun_args).reshape((n_chains, -1))
-        return allreduce(L_x.mean(), op=MPI.SUM)[0] / MPI.COMM_WORLD.Get_size()
+        return allreduce(L_x.mean(), op=MPI.SUM) / MPI.COMM_WORLD.Get_size()
 
     def _expect_fwd(n_chains, log_pdf, expected_fun, pars, x, *expected_fun_args):
         L_x = expected_fun(pars, x, *expected_fun_args)
         L_x_r = L_x.reshape((n_chains, -1))
-        dL_x = allreduce(L_x_r.mean(), op=MPI.SUM)[0] / MPI.COMM_WORLD.Get_size()
+        dL_x = allreduce(L_x_r.mean(), op=MPI.SUM) / MPI.COMM_WORLD.Get_size()
         ΔL_x = L_x - dL_x
         return dL_x, (pars, x, expected_fun_args, ΔL_x)
 
@@ -294,7 +294,7 @@ def test_advanced_jvp():
             term1 = jax.vmap(jnp.multiply)(ΔL_x, log_p)
             term2 = expected_fun(pars, x, *cost_args)
             out = (
-                allreduce(jnp.mean(term1 + term2, axis=0), op=MPI.SUM)[0]
+                allreduce(jnp.mean(term1 + term2, axis=0), op=MPI.SUM)
                 / MPI.COMM_WORLD.Get_size()
             )
             out = out.sum()
